#include "Stellmotor.h"
Stellmotor::Stellmotor(const uint MotorAuf, const uint LED_MotorZu, const uint RT_Regler, 
                 const uint Feedback, const PWM_Status StatusLED )
: _MotorAuf(MotorAuf),
  _MotorZu(LED_MotorZu),
  _RT_Regler(RT_Regler),
  _Feedback(Feedback), 
  _StatusLED(StatusLED)
{
  // Kann sein, dass die Initialisierungen schon alles gemacht haben und der Rumpf hier leer ist
}  

void Stellmotor::init() {                                            // Muss da noch was zurückgegeben werden, falls nur ein Ventil angeschlossen ist?
  pinMode(                _RT_Regler, INPUT);
  pinMode(                _Feedback,  INPUT);
  pinMode(                _MotorAuf,  OUTPUT);
  pinMode(                _MotorZu,   OUTPUT);
  gpio_set_drive_strength(_MotorAuf,  GPIO_DRIVE_STRENGTH_2MA);          // Erspart den Widerstand
  gpio_set_drive_strength(_MotorZu,   GPIO_DRIVE_STRENGTH_2MA);

  _verfahrzeit = kalibrieren();
  _stufenzeit = _verfahrzeit / 9;
  _istwert = 9; // Nach dem kalibrieren ist der Stellmotor auf.
  _StatusLED.Status(_istwert + 1); 
  _fehlerzaehler = 0;
}   
                                                                            
void Stellmotor::init_mess() {                                            // Diese Routine dient dazu, die Paramerter für min/max zu bestimmen
  pinMode(                _RT_Regler, INPUT);
  pinMode(                _Feedback,  INPUT);
  pinMode(                _MotorAuf,  OUTPUT);
  pinMode(                _MotorZu,   OUTPUT);
  gpio_set_drive_strength(_MotorAuf,  GPIO_DRIVE_STRENGTH_2MA);          // Erspart den Widerstand
  gpio_set_drive_strength(_MotorZu,   GPIO_DRIVE_STRENGTH_2MA);

  while (true) {
    _sensorValue = analogRead(_RT_Regler);                               // Sollwert vom Raumthemperaturregler 0...3,85V über Spannungsteiler 330/220 = 2,27V (0...680)
    Serial.print("RT-Fühler-roh:"); Serial.print(_sensorValue); 
    _sollwert =  constrain( _sensorValue, _sensorMin, _sensorMax);       // Falls doch Wert über 700, dann begrenzen
    Serial.print(" RT-Fühler-begrenzt:"); Serial.print(_sollwert); 
    _sollwert =  map(       _sollwert,    _sensorMin, _sensorMax, 0, 9); //Mapping auf Wert von 0 (zu) bis 9 (ganz auf)
    _feedback = analogRead(_Feedback);  
    
    Serial.print(" daraus sollwert:"); Serial.print(_sollwert); 
    Serial.print(" Feedback:"); Serial.println(_feedback); 
    sleep_ms(3000);
  }
}

int Stellmotor::myMap(int x, int in_min, int in_max, int out_min, int out_max) {  
  int in_bereich = in_max - in_min;
  int zwischen =   (x - in_min) * (out_max - out_min) / in_bereich + out_min;  // Kopiert von /home/username/.arduino15/packages/rp2040/hardware/rp2040/2.6.3/ArduinoCore-API/api/common.cpp
  int hysterese =  in_bereich / (out_max - out_min) / 5;                      // Hysterese 20%

  Serial.print("Hysterese:");      Serial.print(hysterese); 
  Serial.print(" Soll (ohne H):");  Serial.print(zwischen); 
  
  if (zwischen == _istwert) return zwischen;                             // Gleiches Ergebnis wie bisher
  if (zwischen == out_max ) return zwischen;                             // Endanschlag wird immer erreicht, sonst Problem mit Ganzzahlarithmetik, dass nur in_max auf out_max gemappt wird
  if (zwischen == out_min ) return zwischen;                             // Endanschlag wird immer erreicht   
  if (zwischen == (_istwert +1))   
     zwischen = (x - hysterese - in_min) * (out_max - out_min) / in_bereich + out_min;  // Hysterese 
  if (zwischen == (_istwert -1))   
     zwischen = (x + hysterese - in_min) * (out_max - out_min) / in_bereich + out_min;  // Hysterese 
  Serial.print(" (mit H):");      Serial.print(zwischen); 
  return zwischen;                                                       // Auch wenn um mehr als einen 0..9 verstellt wird, geht es dirket


}

void Stellmotor::messen() {
  _sensorValue = analogRead(_RT_Regler);                                 // Sollwert vom Raumthemperaturregler 0...3,5V (0...700)
  _sollwert =    constrain( _sensorValue, _sensorMin, _sensorMax);       // Falls doch Wert über 700, dann begrenzen
  _sollwert =    myMap(       _sollwert,    _sensorMin, _sensorMax, 0, 9); //Mapping auf Wert von 0 (zu) bis 9 (ganz auf)
  _feedback =    analogRead(_Feedback);  
  
  Serial.print(" analog:");      Serial.print(  _sensorValue); 
  Serial.print(" Feedback:");   Serial.print(  _feedback); 
  Serial.print(" Grenze_auf:"); Serial.print(  _Feedback_grenze_auf); 
  Serial.print(" Grenze_zu:");  Serial.print(  _Feedback_grenze_zu); 
  Serial.print(" Soll: ");      Serial.print(  _sollwert); 
  Serial.print(" Ist: ");       Serial.print(  _istwert); 
  Serial.print(" Fehler: ");    Serial.println(_fehlerzaehler);  

  if (_sollwert != _istwert) {
    verstellen();
    _StatusLED.Status(_sollwert + 1);                                    // Damit einfacher Blinkimpuls wenn Ventil geschlossen, maximal 10 Impulse bei ganz auf = 9
  }
  if (_fehlerzaehler >= _max_Fehler) init();                             // Wenn  zu viele Fehler, dann neu kalibrieren.
}
    
unsigned long Stellmotor::kalibrieren() {
  unsigned long aufzeit, zuzeit;  
  int feedback_tmp;
  int zeit; //nur für Ausgabe
  
  // Ermitteln des Feedback-Wertes ohne Last durch Motor
  digitalWrite(_MotorZu, LOW); digitalWrite(_MotorAuf, LOW);
  // Jetzt ist der Stellmotor aus
  delay(500);
  _feedback = analogRead(_Feedback);                                    //Aktuell fliest noch kein Strom
  Serial.print("Funktion kalibrieren, ohne Last: "); 
  for (int i = 0; i <= 50; i++) {
    delay(100);
    feedback_tmp = analogRead(_Feedback);
    _feedback = max(_feedback, feedback_tmp);
  }

  _Feedback_grenze_auf = _feedback - _Feedback_delta_auf;                // Meist 900 - 350 = 550
  _Feedback_grenze_zu  = _feedback - _Feedback_delta_zu;                 // Meist 900 - 350 = 550                        
  Serial.print(_feedback); Serial.print("  Grenze_auf: "); Serial.println(_Feedback_grenze_auf);  
  Serial.print(_feedback); Serial.print("  Grenze_zu: ");  Serial.println(_Feedback_grenze_zu);  
  // Erst aufmachen, denn dabei fährt man gegen den "festen" Anschlag, der auch funktioniert, wenn Ventil nicht aufgesteckt ist
  aufzeit = endanschlag(true);                                           // Danach ist der Stellmotor auf
  sleep_ms(2000);                                                        // Damit man auch "hört", dass von auf auf zu gewechselt wird
  zeit = int(aufzeit/1000);
  Serial.print("Aufmachen Zeit: "); Serial.println(zeit);

  // Zumachen
  zuzeit = endanschlag(false);                                           // Danach ist der Stellmotor zu
  sleep_ms(2000);                                                        // Damit man auch "hört", dass von zu auf auf gewechselt wird
  zeit = int(zuzeit/1000);
  Serial.print("Zumachen Zeit: "); Serial.println(zeit);

  // Jetzt wieder Aufmachen
  aufzeit = endanschlag(true);                                           // Danach ist der Stellmotor auf
  sleep_ms(2000);                                                        // Damit man auch "hört", dass von auf auf zu gewechselt wird
  zeit = int(aufzeit/1000);
  Serial.print("Aufmachen Zeit: "); Serial.println(zeit);

  zuzeit = min(zuzeit, aufzeit);
  Serial.print("Minimum Zeit: "); Serial.println(zuzeit);
  return zuzeit;
}

void Stellmotor::verstellen() {
  unsigned long temp_zeit;
  if (_sollwert == 9) {
    temp_zeit = endanschlag(true);
    _istwert = 9;
  }
  else if (_sollwert == 0) {
    temp_zeit = endanschlag(false);
    _istwert = 0;
  }
  else if (_sollwert > _istwert) {
    temp_zeit = (_sollwert - _istwert) * _stufenzeit;
    _istwert = verfahren(true, temp_zeit);
  } 
  else if (_sollwert < _istwert) {
    temp_zeit = (_istwert - _sollwert) * _stufenzeit;
    _istwert = verfahren(false, temp_zeit);
  }

}

int Stellmotor::verfahren(boolean richtung, unsigned long zeit_verfahr) {
  unsigned long zielMillis;  
  int rueckgabewert;

  zielMillis = millis() + zeit_verfahr;
  if (richtung) {
    // TRUE = Aufmachen
    digitalWrite(_MotorZu, LOW); digitalWrite(_MotorAuf, HIGH);
  } else {
    digitalWrite(_MotorZu, HIGH); digitalWrite(_MotorAuf, LOW);
  }

  rueckgabewert = _sollwert;                                             // Wenn keine Ausnahme auftritt, wird dieser Wert zurückgegeben
  Serial.print("Verfahren (Feedback):"); Serial.print(_feedback); Serial.print(" "); 

  while (zielMillis >= millis()) {
    delay(100); // evtl. ganz weg lassen, so sind es 100ms Schritte
    _feedback = analogRead(_Feedback);
    
    Serial.print(_feedback); Serial.print(" "); 

    if (_feedback < _Feedback_grenze_auf) {                              // dieser Fall darf eigentlich nicht eintreten
      if (richtung == true) {
        rueckgabewert = 9;
      } else {
        rueckgabewert = 0;
      }
      _fehlerzaehler++;
      Serial.println(_feedback); 
      Serial.println("!! unerwarteter Endanschlag !!! "); 
      break; 
    }
  }
  digitalWrite(_MotorZu, LOW); digitalWrite(_MotorAuf, LOW);
  Serial.println(_feedback); 
  Serial.println("Reguläres Ende Verfahren"); 
  return rueckgabewert;
}

    
unsigned long Stellmotor::endanschlag(boolean richtung) {
  // true = Aufmachen, false = zumachen
  unsigned long alteMillis, zeit_ms = 0;  
  int feedback_min = 1023, feedback_max = 0, feedback_tmp= 1023;

  alteMillis = millis();
  _feedback = analogRead(_Feedback);
  zeit_ms = millis() - alteMillis;
  if (richtung) {
    // TRUE = Aufmachen
    digitalWrite(_MotorZu, LOW); digitalWrite(_MotorAuf, HIGH);
    Serial.println("Fahrt auf Endanschlag auf: ");
    while ((_feedback >= _Feedback_grenze_auf) || (_feedback < feedback_tmp) || (zeit_ms < 2000)) {          // Ende auch noch nicht, wenn weniger als 2 Sekunden
      if (  _feedback <  _Feedback_grenze_auf) feedback_tmp = _feedback;             // Ende erst, wenn Wert unter Grenze konstant bleibt
      delay(200);
      _feedback = analogRead(_Feedback);
      zeit_ms = millis() - alteMillis;
      feedback_min = min(feedback_min, _feedback);
      feedback_max = max(feedback_max, _feedback);
      Serial.print("Zeit: "); Serial.print(zeit_ms); Serial.print(" FB: "); 
      Serial.print(_feedback); Serial.print(" ("); Serial.print(feedback_min); Serial.print("-");
      Serial.print(feedback_max); Serial.print(") Grenze: "); Serial.println(_Feedback_grenze_auf);
    }
  } else {
    digitalWrite(_MotorZu, HIGH); digitalWrite(_MotorAuf, LOW);
    Serial.println("Fahrt auf Endanschlag zu: ");
    while ((_feedback >= _Feedback_grenze_zu) || (_feedback < feedback_tmp) || (zeit_ms < 2000)) {
      if (  _feedback <  _Feedback_grenze_zu) feedback_tmp = _feedback;             // Ende erst, wenn Wert unter Grenze konstant bleibt
      delay(200);
      _feedback = analogRead(_Feedback);
      zeit_ms = millis() - alteMillis;
      feedback_min = min(feedback_min, _feedback);
      feedback_max = max(feedback_max, _feedback);
      Serial.print("Zeit: "); Serial.print(zeit_ms); Serial.print(" FB: "); 
      Serial.print(_feedback); Serial.print(" ("); Serial.print(feedback_min); Serial.print("-");
      Serial.print(feedback_max); Serial.print(") Grenze: "); Serial.println(_Feedback_grenze_zu);
      // Evtl. muss hier noch eine Schutzfuniktion rein, dass bei nicht richtig aufgestecktem Ventil nach gewisser Zeit abgebrochen wird   !!! TODO
    }
  }

  digitalWrite(_MotorZu, LOW); digitalWrite(_MotorAuf, LOW);
  // Jetzt ist der Stellmotor aus
  
  Serial.print("Endanschlag erreicht nach ms: "); Serial.println(zeit_ms); 
  
  return zeit_ms;
}
